import typing as t

from hcl2.lark_parser import Lark_StandAlone  # type: ignore

from hcl2_ast.ast import Module
from hcl2_ast.transformer import ToAstTransformer


def parse_file(file: t.TextIO) -> Module:
    return parse_string(file.read())


def parse_string(text: str) -> Module:
    return Module(Lark_StandAlone(transformer=ToAstTransformer()).parse(text + "\n"))



import dataclasses
import textwrap
import typing as t

from termcolor import colored

LiteralValue = t.Union[None, bool, int, float, str]


def pformat(value: t.Union[LiteralValue, "Node"]) -> str:
    if isinstance(value, Node):
        return value.pformat()
    if isinstance(value, str):
        return colored(repr(value), "yellow")
    else:
        return colored(repr(value), "cyan")


def pformat_list(values: t.List["Expression"]) -> str:
    result = ""
    if values:
        for value in values:
            result += "\n" + textwrap.indent(pformat(value), "  ") + ","
        result += "\n"
    return result


def indent_but_first_line(text: str, indent: str) -> str:
    lines = text.splitlines()
    if not lines:
        return ""
    return lines[0] + "\n" + textwrap.indent("\n".join(lines[1:]), indent)


class Node:
    """Base class for HCL2 AST nodes."""

    def pformat_field(self, field_name: str) -> str:
        value = getattr(self, field_name)
        if isinstance(value, Expression):
            formatted = value.pformat()
        elif isinstance(value, list):
            formatted = f"[{pformat_list(value)}]"
        else:
            formatted = pformat(value)
        return formatted

    def pformat(self) -> str:
        args = []
        fields = dataclasses.fields(self)
        for field in fields:
            args.append(colored(field.name, "cyan") + "=" + self.pformat_field(field.name))
        if any("\n" in arg for arg in args) and len(args) > 1:
            sep = ",\n"
            args_string = f'(\n{sep.join(textwrap.indent(arg, "  ") for arg in args)}\n)'
        else:
            args_string = f'({", ".join(args)})'
        return f"{colored(type(self).__name__, 'blue')}{args_string}"


class Expression(Node):
    """Base class for nodes that represent expressions in HCL2."""


@dataclasses.dataclass
class Literal(Expression):
    value: LiteralValue

    def __post_init__(self) -> None:
        assert isinstance(self.value, (type(None), bool, int, float, str)), self.value
        if "Identifier" in str(self.value):
            import pdb

            pdb.set_trace()


@dataclasses.dataclass
class Array(Expression):
    values: t.List[Expression]


@dataclasses.dataclass
class Object(Expression):
    fields: t.Dict[str, Expression]

    def pformat_field(self, field_name: str) -> str:
        if field_name == "fields":
            if not self.fields:
                return "{}"
            result = "{"
            for key, value in self.fields.items():
                result += f'\n  {pformat(key)}: {indent_but_first_line(value.pformat(), "  ")}'
            return result + "}"
        return super().pformat_field(field_name)


@dataclasses.dataclass
class FunctionCall(Expression):
    name: str
    args: t.List[Expression]

    def __post_init__(self) -> None:
        assert all(isinstance(arg, Expression) for arg in self.args), self.args


@dataclasses.dataclass
class Identifier(Expression):
    name: str


@dataclasses.dataclass
class GetAttr(Expression):
    on: t.Optional["Expression"]
    name: str


@dataclasses.dataclass
class UnaryOp(Expression):
    op: t.Literal["-", "!"]
    expr: Expression


@dataclasses.dataclass
class BinaryOp(Expression):
    op: t.Literal["==", "!=", "<", ">", "<=", ">=", "-", "*", "/", "%", "&&", "||", "+"]
    left: Expression
    right: Expression


class Stmt(Node):
    """Base class for nodes that represent statements in HCL2."""


@dataclasses.dataclass
class Attribute(Stmt):
    key: str
    value: Expression


@dataclasses.dataclass
class Block(Stmt):
    name: str
    args: t.List[Expression]
    body: t.List[Stmt]


@dataclasses.dataclass
class Module(Node):
    body: t.List[Stmt]



import argparse

from hcl2_ast.api import parse_file

parser = argparse.ArgumentParser()
parser.add_argument("file", type=argparse.FileType("r"))


def main() -> None:
    args = parser.parse_args()
    module = parse_file(args.file)
    print(module.pformat())


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['hcl2_ast']

package_data = \
{'': ['*']}

install_requires = \
['python-hcl2>=3.0.5,<4.0.0', 'termcolor>=1.1.0,<2.0.0']

setup_kwargs = {
    'name': 'hcl2-ast',
    'version': '0.1.1',
    'description': '',
    'long_description': '# hcl2-ast\n\nA [HCL2][] parser based on [python-hcl2][] that produces an Abstract Syntax Tree.\n\n  [HCL2]: https://github.com/hashicorp/hcl/blob/main/README.md\n  [python-hcl2]: https://pypi.org/project/python-hcl2/\n\n> __Note__: This project is in an early stage. It does not currently cover all HCL2 syntax features\n> and does not have good test coverage.\n\n## Usage\n\n```py\nfrom hcl2_ast import parse_string\n\nmodule = parse_string("""\n  task "say-hello" {\n    to = "World"\n  }\n""")\n\nprint(module.pformat())\n```\n\nOutputs:\n\n```py\nModule(body=[\n  Block(\n    name=\'task\',\n    args=[\n      \'say-hello\',\n    ],\n    body=[\n      Attribute(key=\'to\', value=Literal(value=\'World\')),\n    ]\n  ),\n])\n```\n\n## Compatibility\n\nhcl2-ast requires Python 3.6 or higher.\n\n## Known issues\n\n* No understanding of operator precedence in expressions (grouping with parentheses works as expected)\n',
    'author': 'Niklas Rosenstein',
    'author_email': 'rosensteinniklas@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



""" Uses python-hcl2 to parse a Kraken configuration file, but outputs it in a different format that instead
retains the order of assignments and blocks. """

import typing as t

from hcl2.transformer import DictTransformer  # type: ignore[import]
from lark import Token

from hcl2_ast.ast import (
    Array,
    Attribute,
    BinaryOp,
    Block,
    Expression,
    FunctionCall,
    GetAttr,
    Identifier,
    Literal,
    Object,
    UnaryOp,
)


class ToAstTransformer(DictTransformer):  # type: ignore[misc]
    def to_expression(self, value: t.Any) -> Expression:
        if isinstance(value, Expression):
            return value
        elif type(value) == str:
            return Literal(value)
        else:
            raise TypeError(value)

    # DictTransformer

    def unary_op(self, args: t.List[t.Any]) -> str:
        return UnaryOp(str(args[0]), args[1])

    def binary_term(self, args: t.List[t.Any]) -> t.List[t.Any]:
        return args

    def binary_op(self, args: t.List[t.Any]) -> str:
        assert len(args) == 2
        return BinaryOp(str(args[1][0]), args[0], args[1][1])

    def get_attr(self, args: t.List[t.Any]) -> str:
        return self.to_expression(args[0])

    def get_attr_expr_term(self, args: t.List[t.Any]) -> str:
        assert len(args) == 2, args
        if isinstance(args[0], Expression):
            on = args[0]
        else:
            on = Identifier(args[0])
        return GetAttr(on, args[1].name)

    def float_lit(self, args: t.List[t.Any]) -> Literal:
        return Literal(super().float_lit(args))

    def int_lit(self, args: t.List[t.Any]) -> Literal:
        return Literal(super().int_lit(args))

    def expr_term(self, args: t.List[t.Any]) -> t.Union[Literal, t.Any]:
        if isinstance(args[0], Token) and args[0].type == "STRING_LIT":
            return Literal(self.strip_quotes(args[0].value))
        return super().expr_term(args)

    def identifier(self, value: t.List[Token]) -> Identifier:
        # Making identifier a token by capitalizing it to IDENTIFIER
        # seems to return a token object instead of the str
        # So treat it like a regular rule
        # In this case we just convert the whole thing to a string
        if value[0].value == "null":
            return Literal(None)
        if value[0].value == "true":
            return Literal(True)
        if value[0].value == "false":
            return Literal(False)
        return Identifier(str(value[0]))

    def attribute(self, args: t.List[t.Any]) -> Attribute:
        return Attribute(args[0].name, self.to_expression(args[1]))

    def function_call(self, args: t.List[t.Any]) -> FunctionCall:
        args = self.strip_new_line_tokens(args)
        assert len(args) in (1, 2), args
        if len(args) == 1:
            args.append([])
        args[1] = [self.to_expression(self.strip_quotes(arg)) for arg in args[1]]
        return FunctionCall(args[0], args[1])

    def body(self, args: t.List[t.Any]) -> t.List[t.Any]:
        return args

    def block(self, args: t.List[t.Any]) -> Block:
        args = self.strip_new_line_tokens(args)

        # if the last token is a string instead of an object then the block is empty
        # such as 'foo "bar" "baz" {}'
        # in that case append an empty object
        if isinstance(args[-1], str):
            args.append({})

        args = [self.to_expression(self.strip_quotes(arg)) for arg in args[:-1]] + [args[-1]]

        return Block(args[0].name, args[1:-1], args[-1])

    def object(self, args: t.List[t.Any]) -> Object:
        args = self.strip_new_line_tokens(args)
        result: t.Dict[str, t.Any] = {}
        for arg in args:
            result.update(arg)
        return Object(result)

    def tuple(self, args: t.List[t.Any]) -> Array:
        return Array(super().tuple(args))
